// Copyright (c) 2000-2011 ActiveState Software Inc.
// See the file LICENSE.txt for licensing information.

/**
 *
 * stackato.js -- JS code for the Stackato dashboard
 *
 */

if (typeof(ko)=='undefined') {
    var ko = {};
}
if (typeof(ko.stackato)=='undefined') {
    ko.stackato = {};
}
var widgets = {};
var _stackatoWindow = window;
(function() {

var bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
    .getService(Components.interfaces.nsIStringBundleService)
    .createBundle("chrome://stackatotools/locale/stackato.properties");
var log;

var gko = null;
var gWindow = null;
var gStackatoPrefs;

var firstTimeRun = false;

this.focus = function() {
    gWindow.focus();
    setTimeout(function() {
        _stackatoWindow.focus();
    }, 100);
}

this.onLoad = function() {
    gWindow = window.arguments[0].window;
    gko = window.arguments[0].ko;
    log = gko.logging.getLogger("stackato");
    setTimeout(function(this_) {
        // show the window first, then set things up.
        try {
            this_.initialize();
            this_.getAllData();
        } catch(ex) {
            log.exception(ex, "Error in onLoad: ");
        }
    }, 1000, this);
};

this.initialize = function() {
    this.treeNames = ["mainTree", "servicesProvisionedTree",
                      "servicesSystemTree", "frameworksTree", "runtimesTree", "targetsTree"];
    this.treeChildrenNames = ["mainTree_treechildren",
                           "servicesProvisionedTree_treechildren",
                           "servicesSystemTree_treechildren",
                           "frameworksTree_treechildren",
                           "runtimesTree_treechildren",
                           "targetsTree_treechildren"
                           ];
    this.stackatoService = Components.classes["@activestate.com/koStackatoServices;1"]
            .createInstance(Components.interfaces.koIStackatoServices);
    this.stackatoService.initialize();
    this.jsonService = Components.classes["@mozilla.org/dom/json;1"]
    .createInstance(Components.interfaces.nsIJSON);
    
    widgets.fields = {};
    
    for each (var name in [
        "target_textbox", "change_target_button", "revert_target_button",
        "user_logged_in", "user1_textbox", "change_user1_button",
        "user_logged_out", "user2_textbox", "user2_password", "change_user2_button"
    ]) {
        widgets.fields[name] = document.getElementById(name);
    }
    
    widgets.trees = {};
    widgets.views = {};
    for each (var name in this.treeNames) {
        widgets[name] = widgets.trees[name] = document.getElementById(name);
    }
    this.servicesTreeManager = new ko.stackato.trees.ServicesTreeManager();

    this.setupNodeControls();
    this.user = this.password = "";
    this.runtimeNames = [];
    this.provisionedServiceNames = [];
    this.systemServiceNames = [];
    this.targetNames = [],
    this.hasLoggedOut = false;
    this.pendingRequests = {};

    this.finishBuildingToolbar("stackato_apps_toolbar");
    
    var nsIDOMKeyEvent = Components.interfaces.nsIDOMKeyEvent;
    this.arrowKeys = [nsIDOMKeyEvent.DOM_VK_UP,
                      nsIDOMKeyEvent.DOM_VK_DOWN,
                      nsIDOMKeyEvent.DOM_VK_LEFT,
                      nsIDOMKeyEvent.DOM_VK_RIGHT];

    if (!gko.prefs.hasPref("stackato")) {
        firstTimeRun = true;
        gStackatoPrefs = Components.classes["@activestate.com/koPreferenceSet;1"].createInstance();
        gko.prefs.setPref("stackato", gStackatoPrefs);
        gStackatoPrefs.setLongPref("version", 1);
    } else {
        gStackatoPrefs = gko.prefs.getPref("stackato");
        if (!gStackatoPrefs.hasPref("notFirstTime")
            || !gStackatoPrefs.getBooleanPref("notFirstTime")) {
            firstTimeRun = true;
        }
    }
    if (!gStackatoPrefs.hasPref("apps")) {
        gStackatoPrefs.setPref("apps", Components.classes["@activestate.com/koPreferenceSet;1"].createInstance());
    }
    if (!gStackatoPrefs.hasPref("boxIsClosed")) {
        gStackatoPrefs.setPref("boxIsClosed", Components.classes["@activestate.com/koPreferenceSet;1"].createInstance());
    } else {
        var boxIsClosedPrefs = gStackatoPrefs.getPref("boxIsClosed");
        var ids = {};
        boxIsClosedPrefs.getPrefIds(ids, {});
        for each (var boxId in ids.value) {
                var parts = this.jsonService.decode(boxIsClosedPrefs.getStringPref(boxId));
                var imageNode = document.getElementById(parts[0]);
                var vboxNode = document.getElementById(parts[1]);
                imageNode.setAttribute("class", "closebox-icon");
                this.toggleBox(imageNode, vboxNode, parts[2], true);
            }
    }
    this.showEnvironmentVariables = false;                
                
    //TODO: Prevent sorting in the main tree.  The problem is that the
    // tree.xml handler handles the eventPhase = atTarget value, which
    // addEventListener doesn't handle, and an attempt to set up a binding failed.
    
    //var mainTreecols = Array.slice(widgets.trees.mainTree.getElementsByTagName("treecol"));
    //// The clicks are handled at phase="target"; so set useCapture to false
    //for each (var treecol in mainTreecols) {
    //    treecol.addEventListener("click", this.handleMainTreecolClick, true);
    //    treecol.addEventListener("click", this.handleMainTreecolClick, false);
    //}
};

//this.handleMainTreecolClick = function(event) {
//    // squelch sort.
//    event.stopPropagation();
//    event.preventDefault();
//};

this._attributesToCopy = ['oncommand', 'disableUnless', 'disableIf'];

this.finishBuildingToolbar = function(toolbar_id) {
    var toolbar = document.getElementById(toolbar_id);
    var childNodes = toolbar.childNodes;
    var i, j, targetNode, srcNode, srcId, attributes, attrLen, attrNode;
    for (i = 0; i < childNodes.length; i++) {
        targetNode = childNodes[i];
        srcId = targetNode.getAttribute("inheritFrom");
        if (!srcId || !(srcNode = document.getElementById(srcId))) {
            continue;
        }
        attributes = srcNode.attributes;
        attrLen = attributes.length;
        for (j = 0; j < attrLen; ++j) {
            attrNode = attributes[j];
            if (this._attributesToCopy.indexOf(attrNode.name) >= 0) {
                targetNode.setAttribute(attrNode.name, attrNode.value);
            }
        }
    }
};    

this.updateTarget = function() {
    var newTarget = widgets.fields.target_textbox.value;
    if (newTarget) {
        this.stackatoService.target = this._target = newTarget;
        this._updateTargetButtons(true);
    }
};

this.revertTarget = function(target) {
    if (typeof(target) == "undefined") target = null;
    if (target) {
        this._target = target;
    }
    widgets.fields.target_textbox.value = this._target;
    this._updateTargetButtons(true);
};

this.refreshUser = function(userName) {
    var user = userName;
    if (user) {
        this._updateUserBoxes(true);
        widgets.fields["user1_textbox"].value = user;
    } else {
        this._updateUserBoxes(false);
        widgets.fields["user2_textbox"].value = this.user;
        widgets.fields["user2_password"].value = this.password;
        this.checkLogin();
    }
};

this._tf = ["false", "true"];

this._updateUserBoxes = function(isLoggedIn) {
    var index = isLoggedIn ? 0 : 1;
    widgets.fields["user_logged_in"].setAttribute("hidden", this._tf[index]);
    widgets.fields["user_logged_out"].setAttribute("hidden", this._tf[1 - index]);
};

this.login = function() {
    var user = widgets.fields["user2_textbox"].value;
    if (!user) {
        alert("no user supplied");
        return;
    }
    var password = widgets.fields["user2_password"].value;
    var this_ = this;
    this.hasLoggedOut = false;
    var handler = {
      setData: function(data) {
            // Hardwired string from Stackato client, do not i18nize
            if (data && data.indexOf("Successfully logged into") == 0) {
                this_.user = user;
                this_.password = password;
                // Do other things on login....
                this_._updateUserBoxes(!!user);
                if (user) {
                    widgets.fields["user1_textbox"].value = user;
                    this_.setupApplicationsTree(true);
                } else {
                    this_._updateUserBoxes(false);
                }
            } else {
                // Hardwired string from Stackato client, do not i18nize
                if (data && data.indexOf("Problem with login, invalid account or password") == 0) {
                    ko.dialogs.alert(data);
                }
                this_._updateUserBoxes(false);
            }
        }
    };
    this.wrapCallbackFunction("login", "user-icon",
                              handler,
                              null, false, [user, password]);
};

this.logout = function() {
    var this_ = this;
    var user = widgets.fields.user1_textbox.value;
    var handler = {
      setData: function(data) {
            this_.hasLoggedOut = true;
            this_.refreshUser(null);
            // Hardwired string from Stackato client, do not i18nize
            if (data && data.indexOf("Successfully logged out of") == 0) {
                // Do other things on logout....
                if (user) {
                    widgets.fields["user2_textbox"].value = user || this_.user;
                    if (this_.password) {
                        widgets.fields["user2_password"].value = this_.password;
                    }
                    this_.clearTrees();
                }
            }
        }
    };
    this.wrapCallbackFunction("logout", "user-icon",
                              handler,
                              null,
                              /*useJSON=*/false);
};

this.checkLogin = function() {
    widgets.fields.change_user2_button.disabled = (
        !widgets.fields.user2_textbox.value
        || !widgets.fields.user2_password.value);
    
};
this.checkUpdateTarget = function(menulist) {
    var newTarget = widgets.fields.target_textbox.value;
    this._updateTargetButtons(!newTarget || newTarget == this._target);
};

this.changeTarget = function(menulist) {
    log.debug("In ko.stackato.changeTarget");
}

this._updateTargetButtons = function(disabledStatus) {
    widgets.fields.change_target_button.disabled = disabledStatus;
    widgets.fields.revert_target_button.disabled = disabledStatus;
};

this._addStyle = function _addStyle(node, value) {
    var style = node.getAttribute("style");
    if (style) {
        style += " " + value;
    } else {
        style = value;
    }
    node.setAttribute("style", value);
}

this.toggleBox = function toggleBox(imageNode, vboxNode, otherNodeNames,
                                    skipPrefsPart) {
    if (typeof(otherNodeNames) == "undefined") otherNodeNames = [];
    if (typeof(skipPrefsPart) == "undefined") skipPrefsPart = false;
    var doClose = imageNode.getAttribute('class') == "closebox-icon";
    if (!skipPrefsPart) {
        var boxIsClosedPrefs = gStackatoPrefs.getPref("boxIsClosed");
        if (doClose) {
            var names = [imageNode.id, vboxNode.id, otherNodeNames];
            var namesAsJSON = this.jsonService.encode(names);
            boxIsClosedPrefs.setStringPref(imageNode.id, namesAsJSON);
        } else if (boxIsClosedPrefs.hasPref(imageNode.id)) {
            boxIsClosedPrefs.deletePref(imageNode.id);
        }
    }
    var targetNodes = vboxNode.getElementsByTagName("tree");
    var targetNodeArray = [];
    for (var i = 0; i < targetNodes.length; i++) {
        targetNodeArray.push(targetNodes[i]);
    }
    for (i = 0; i < otherNodeNames.length; i++) {
        var node = document.getElementById(otherNodeNames[i]);
        if (node) {
            targetNodeArray.push(node);
        }
    }
    if (doClose) {
        this.closeToggleBox(vboxNode, targetNodeArray, imageNode);
    } else {
        this.openToggleBox(vboxNode, targetNodeArray, imageNode);
    }
};

this.setSquirreledFlex = function(node) {
    if (node.hasAttribute("koSquirreledFlex")) {
        node.setAttribute("flex",
                          node.getAttribute("koSquirreledFlex"));
    }
};

this.zeroAndSquirrelFlex = function(node) {
    var flex = node.getAttribute("flex");
    if (flex) {
        node.setAttribute("koSquirreledFlex", flex);
        node.setAttribute("flex", "0");
    }
};

this.closeToggleBox = function(vboxNode, targetNodes, imageNode) {
    for each (var targetNode in targetNodes) {
        this._addStyle(targetNode, 'visibility: collapse;');
        this.zeroAndSquirrelFlex(targetNode);
    }
    this.zeroAndSquirrelFlex(vboxNode);
    imageNode.setAttribute("tooltiptext", "show pane");
    imageNode.setAttribute('class', "openbox-icon");
};

this.openToggleBox = function(vboxNode, targetNodes, imageNode) {
    for each (var targetNode in targetNodes) {
        this._addStyle(targetNode, 'visibility: visible;');
        this.setSquirreledFlex(targetNode);
    }
    this.setSquirreledFlex(vboxNode);
    imageNode.setAttribute("tooltiptext", "hide pane");
    imageNode.setAttribute('class', "closebox-icon");
};

this.showTreeColumn = function(node) {
    node.removeAttribute('hidden');
}

this.hideTreeColumn = function(node) {
    node.setAttribute('hidden', 'true');
};

this.toggleByName = function(names, isChecked) {
    var func = isChecked ? this.showTreeColumn : this.hideTreeColumn;
    names.forEach(function(name) {
        func(document.getElementById(name));
    });
};

this.toggleShowEnvironment = function(checkbox) {
    var checked = checkbox.checked;
    this.showEnvironmentVariables = checked;
    gStackatoPrefs.setBooleanPref("showEnvironment", checked);
    this.doToggleShowEnvironment(checked);
}

this.getAndUpdateEnvironmentVarsForAppname = function(view, appName, callback) {
    var handler = {
      onError: function() { log.debug("env appName failed"); } ,
      onStopped: function(){ log.debug("env appName was stopped"); },
      setData: function(results) {
          view.addEnvironmentData(appName, results);
      }
    };
    this.wrapCallbackFunction("runCommand",
                              "toggle_applications_button",
                              handler,
                              callback,
                              true, /*useJSON */
                              ["env", appName, "--json"]);
};

this.doToggleShowEnvironment = function(checked, callback) {
    var names = ["post_url", "env_name", "post_env_name", "env_value"];
    if (typeof(callback) == "undefined") callback = null;
    var view = widgets.views.mainTree;
    view.setEnvironmentValueStatus(checked);
    if (!checked) {
        view.removeEnvValuesFromTree();
        this.toggleByName(names, false);
        if (callback) callback();
    } else {
        var appNames = view.getNames();
        var appName;
        var this_ = this;
        var updateFunc = function(i) {
            if (i >= appNames.length) {
                this_.toggleByName(names, true);
                this_.focus();
                if (callback) callback();
            } else {
                this_.getAndUpdateEnvironmentVarsForAppname(view, appNames[i],
                                                            function() {
                                                                updateFunc(i + 1)
                                                                    });
            }
        }
        updateFunc(0);
    }
    
};

this.toggleShowStatistics = function(checkbox) {
    var names = ["post_env_value", "stats_instance_num", "post_stats_instance_num",
                 "stats_cpu_cores", "post_stats_cpu_cores", "stats_memory_limit",
                 "post_stats_memory_limit", "stats_disk_limit", "post_stats_disk_limit",
                 "stats_uptime", "post_stats_uptime"
                 ];
    var checked = checkbox.checked;
    widgets.views.mainTree.showStatistics = checked;
    this.toggleByName(names, checked);
};

this.toggleShowPassword = function(checkbox, password_id) {
    var checked = checkbox.checked;
    var textbox = document.getElementById(password_id);
    if (checked) {
        textbox.removeAttribute('type');
    } else {
        textbox.setAttribute('type', 'password');
    }
};

this.toggleDeckPanel = function(checkbox) {
    document.getElementById("stackato_deck").selectedIndex = checkbox.checked ? 1 : 0;
    //var deck = document.getElementById("stackato_deck");
    //deck.selectedIndex = checkbox.checked ? 1 : 0;
    //if (checked) {
    //    textbox.removeAttribute('type');
    //} else {
    //    textbox.setAttribute('type', 'password');
    //}
};

this.showTargetInfo = function() {
    this.doApplicationCommand(["info"], null);
};

this.visitStackatoHelpPage = function() {
    gko.browse.openUrlInDefaultBrowser("http://docs.stackato.com/");
};

this.visitStackatoHomePage = function() {
    gko.browse.openUrlInDefaultBrowser("http://www.activestate.com/cloud");
};

this.wrapCallbackFunction = function(methodName,
                                     toggleButtonID,
                                     dataHandler,
                                     nextFunc,
                                     useJSON,
                                     args) {
    var this_ = this;
    if (typeof(nextFunc) == "undefined") nextFunc = null;
    if (typeof(useJSON) == "undefined") useJSON = true;
    if (typeof(args) == "undefined") args = [];
    var toggleButtonImage, prevClass = null;
    if (toggleButtonID) {
        toggleButtonImage = document.getElementById(toggleButtonID);
        prevClass = toggleButtonImage.getAttribute("class");
        toggleButtonImage.setAttribute("class", "async_operation");
    }
    var async_callback = {
        callback: function(result, data) {
            if (this_.hasLoggedOut && methodName != "logout") {
                return;
            }
            if (toggleButtonID) {
                toggleButtonImage.setAttribute("class", prevClass);
            }
            delete this_.pendingRequests[methodName];
            if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                if (dataHandler.onError) {
                    dataHandler.onError(data && data.stderr);
                } else if (!data || !data.stderr) {
                    ko.dialogs.alert(bundle.formatStringFromName("Stackato X failed", [methodName], 1));
                } else {
                    ko.dialogs.alert(bundle.formatStringFromName("Stackato X ERROR", [methodName, data.stderr], 2));
                }
            } else if (result == Components.interfaces.koIAsyncCallback.RESULT_STOPPED) {
                if (dataHandler.onStopped) {
                    dataHandler.onStopped();
                } else {
                    ko.dialogs.alert(bundle.formatStringFromName("Stackato X operation was stopped", [methodName], 1));
                }
            } else if (data.stderr) {
                if (dataHandler.onError) {
                    dataHandler.onError(data.stderr);
                } else {
                    ko.dialogs.alert(bundle.formatStringFromName("Stackato X ERROR", [methodName, data.stderr], 2));
                }
            } else {
                var processedData = (useJSON
                                     ? this_.jsonService.decode(data.stdout)
                                     : data.stdout);
                dataHandler.setData(processedData);
                if (nextFunc) {
                    nextFunc.call(this_);
                }
            }
        }
    };
    if (methodName == "runCommand") {
        return this.pendingRequests[methodName] = this.stackatoService.runCommand(async_callback, args.length, args);
    } else {
        // grumble must be a better way to do this
        switch (args.length) {
            case 0:
            return this.pendingRequests[methodName] = this.stackatoService[methodName](async_callback);
            case 1:
            return this.pendingRequests[methodName] = this.stackatoService[methodName](args[0], async_callback);
            case 2:
            return this.pendingRequests[methodName] = this.stackatoService[methodName](args[0], args[1], async_callback);
            default:
            log.debug("internal error: have to handle " + args.length + " args");
        }
    }
    return null;
};

this.handleFirstTimeRun = function() {
    if (firstTimeRun) {
        var _target_textbox_orig = document.getElementById("target_textbox_orig");
        _target_textbox_orig.value = bundle.GetStringFromName("Enter a Stackato target name here");
        var func = function(event) {
            _target_textbox_orig.value = "";
            _target_textbox_orig.removeEventListener("focus", func, false);
            gStackatoPrefs.setBooleanPref("notFirstTime", true);
        };
        _target_textbox_orig.addEventListener("focus", func, false);
    }
};

this.getAllData = function() {
    var this_ = this;
    var handleTarget = {
      onError: function(data) {
            this_.handleFirstTimeRun();
        },
      setData: function(data) {
            var target = data.target;
            if (target) {
                this_.revertTarget(target);
                this_.goGetUser();
                if (firstTimeRun) {
                    firstTimeRun = false;
                    gStackatoPrefs.setBooleanPref("notFirstTime", true);
                }
            } else {
                this_.handleFirstTimeRun();
                // Don't get any more data
            }
        }
    };
    this.wrapCallbackFunction("getCurrentTarget", "target-icon",
                              handleTarget);
};

this.goGetUser = function() {
    var this_ = this;
    var handleUser = {
      setData: function(data) {
            var user = (data && data != "N/A") ? data : "";
            this.user = user;
            this_.refreshUser(user);
            if (user) {
                    this_.setupApplicationsTree(true);
            }
        },
      onError: function(msg) {
            ko.dialogs.alert(bundle.formatStringFromName("Stackato getCurrentUser MSG", [msg], 1));
            // Make sure the Login thing is visible
            this_.refreshUser(false);
        },
    };
    this.wrapCallbackFunction("getCurrentUser", "user-icon",
                              handleUser);
};

this._dbExplorerDatabases = {
    "mysql": "MySQL",
    "postgresql": "PostgreSQL",
    "oracle": "Oracle"
};
    
this.getDBVendor = function(dbServiceName) {
    var view = widgets.views.servicesProvisionedTree;
    if (!view) return null; // happens at startup
    var index = view.getIndexForName(dbServiceName);
    if (index == -1) {
        return null;
    }
    return view.getRow(index).fields.servicesProvisionedTree_vendor;
};

this.canonicalizeDBTypeName = function(dbTypeName) {
    var res = this._dbExplorerDatabases[dbTypeName.toLowerCase()];
    return res || dbTypeName;
};

this.handleDatabase = function(dbServiceName) {
    var dbVendorName = this.getDBVendor(dbServiceName);
    return !dbVendorName || !!this._dbExplorerDatabases[dbVendorName.toLowerCase()];
};

this.getDatabaseInfo = function(view, callback) {
    /**
     * @view {appTree view}
     */
    var appNames = view.getNames();
    var currentAppName = null;
    var appNameDBNameSets = []; // array of:
    // { appName:name, dbName: db }
    var fields, row, info, numRows = view.rowCount;
    var currentAppRow;
    for (var i = 0; i < numRows; i++) {
        var fields = view.getRow(i).fields;
        if (fields.name) {
            currentAppRow = i;
            currentAppName = fields.name;
        }
        if (fields.serviceName) {
            info = {
              appName: currentAppName,
              dbName: fields.serviceName
            };
            appNameDBNameSets.push(info);
        }
    }
    var this_ = this;
    this.databasesByAppName = {};
    // map appName => { serviceName => {info}}
    // info: hash of { databaseTypeName, host, port, user, passwd, dbname }
    var idx;
    var handler = {
      setData: function(data) {
            // stuff data into databasesByAppName
            if (!this_.databasesByAppName[info.appName]) {
                this_.databasesByAppName[info.appName] = {};
            }
            var thisInfoH = this_.databasesByAppName[info.appName];
            var thisInfo = {
              databaseTypeName: this_.canonicalizeDBTypeName(data['shell']),
            };
            thisInfoH[info.dbName] = thisInfo;
            var args = data['args'];
            var lim = args.length;
            var arg;
            for (var i = 0; i < lim; i++) {
                arg = args[i];
                if (arg == '-h' || arg == '--host') {
                    thisInfo['host'] = args[i + 1];
                    i += 1;
                } else if (arg == '--port') {
                    thisInfo['port'] = args[i + 1];
                    i += 1;
                } else if (arg == '-p') {
                    thisInfo['passwd'] = args[i + 1];
                    i += 1;
                } else if (arg.substr(0, 2) == '-p') {
                    thisInfo['passwd'] = arg.substr(2);
                } else if (arg == '-u') {
                    thisInfo['user'] = args[i + 1];
                    i += 1;
                } else if (arg.substr(0, 2) == '-u' && arg.length > 2) {
                    thisInfo['user'] = arg.substr(2);
                } else {
                    thisInfo['dbName'] = arg;
                }
            }
            idx += 1;
        }
    };
    var getDatabaseInfoForApp = function() {
        if (idx >= appNameDBNameSets.length) {
            if (callback) callback();
            return;
        }
        info = appNameDBNameSets[idx];
        this_.wrapCallbackFunction("runCommand",
                                   "toggle_applications_button",
                                   handler,
                                   getDatabaseInfoForApp,
                                   true, /*useJSON */
                                   ["dbshell", info.appName, info.dbName, "--json"]);
    };
    idx = 0;
    getDatabaseInfoForApp();
};

this.setupApplicationsTree = function(doFullCallback) {
    var view;
    if (widgets.views.mainTree) {
        view = widgets.views.mainTree;
    } else {
        view = widgets.views.mainTree = new ko.stackato.trees.ApplicationsTree(gStackatoPrefs);
        widgets.trees.mainTree.treeBoxObject.view = view;
    }
    var this_ = this;
    var nextCallback = function() {
        this_.getDatabaseInfo(view, (doFullCallback
                                     ? function() {this_.setupServicesTree();}
                                     : null));
    };
    var callback = (gStackatoPrefs.hasBooleanPref("showEnvironment")
                    && gStackatoPrefs.getBooleanPref("showEnvironment")
                    ? function() {
                        document.getElementById("maintree_show_env").checked = true;
                        this_.showEnvironmentVariables = true;
                        this_.doToggleShowEnvironment(true, nextCallback);
                    } : nextCallback);
    this.wrapCallbackFunction("getApplications",
                              "toggle_applications_button",
                              view,
                              callback);
};

this.setupServicesTree = function(inStartup) {
    var view = widgets.views.servicesProvisionedTree = this.servicesTreeManager.servicesProvisionedTree;
    widgets.trees.servicesProvisionedTree.treeBoxObject.view = view;
    
    var view2 = widgets.views.servicesSystemTree = this.servicesTreeManager.servicesSystemTree;
    widgets.trees.servicesSystemTree.treeBoxObject.view = view2;
    var this_ = this;
    var callback = function() {
        this_.provisionedServiceNames = view.getNames();
        this_.systemServiceNames = view2.getNames();
        this_.setupFrameworksTree();
    };
    this.wrapCallbackFunction("getServices",
                              "toggle_provisioned_services_button",
                              this.servicesTreeManager,
                              callback);
};

this.refreshProvisionedServicesTree = function(outer_callback, iconID) {
    var this_ = this;
    var callback = function() {
        this_.provisionedServiceNames =
                widgets.views.servicesProvisionedTree.getNames();
        if (outer_callback) {
            outer_callback();
        }
    };
    this.wrapCallbackFunction("getServices",
                              iconID,
                              this.servicesTreeManager,
                              callback);
};

this.setupFrameworksTree = function() {
    var view = widgets.views.frameworksTree = new ko.stackato.trees.FrameworksTree();
    widgets.trees.frameworksTree.treeBoxObject.view = view;
    this.wrapCallbackFunction("getFrameworks",
                              "toggle_frameworks_runtimes_button",
                              view,
                              this.setupRuntimesTree);
};

this.setupRuntimesTree = function() {
    var view = widgets.views.runtimesTree = new ko.stackato.trees.RuntimesTree();
    widgets.trees.runtimesTree.treeBoxObject.view = view;
    var this_ = this;
    var callback = function() {
        this_.runtimeNames = view.getRuntimeNames();
        this_.setupTargetsTree();
    };
    this.wrapCallbackFunction("getRuntimes",
                              "toggle_frameworks_runtimes_button",
                              view,
                              callback);
};

this.setupTargetsTree = function() {
    var view = widgets.views.targetsTree = new ko.stackato.trees.TargetsTree();
    widgets.trees.targetsTree.treeBoxObject.view = view;
    var this_ = this;
    var callback = function() {
        this_.targetNames = view.getTargetNames();
        this_.updateTargetsMenu(this_.targetNames);
    };
    this.wrapCallbackFunction("getTargets",
                              "toggle_targets_button",
                              view,
                              callback);
};

this.setupNodeControls = function() {
    var nodeControls = [
        {imageNodeName: "toggle_applications_button",
         vboxNodeName: "vbox_applications",
         otherNodeNames: [],
        },
        {imageNodeName: "toggle_provisioned_services_button",
         vboxNodeName: "vbox_provisioned_services",
         otherNodeNames: ['hbox_provisioned_services_caption'],
        },
        {imageNodeName: "toggle_system_services_button",
         vboxNodeName: "vbox_system_services",
         otherNodeNames: ['hbox_system_services_caption'],
        },
        {imageNodeName: "toggle_frameworks_runtimes_button",
         vboxNodeName: "vbox_frameworks_runtimes",
         otherNodeNames: ["vbox_frameworks", "vbox_runtimes"],
        },
        {imageNodeName: "toggle_targets_button",
         vboxNodeName: "vbox_targets",
         otherNodeNames: [],
        }
    ];
    for each (var nodeControl in nodeControls) {
        var imageNode = document.getElementById(nodeControl.imageNodeName);
        var vboxNode = document.getElementById(nodeControl.vboxNodeName);
        var treeNodesObj = vboxNode.getElementsByTagName("tree");
        var otherNodeNames = nodeControl.otherNodeNames;
        var targetNodeArray = [];
        for (var i = 0; i < treeNodesObj.length; i++) {
            targetNodeArray.push(treeNodesObj[i]);
        }
        for (i = 0; i < otherNodeNames.length; i++) {
            var node = document.getElementById(otherNodeNames[i]);
            if (node) {
                targetNodeArray.push(node);
            }
        }
        var doClose = imageNode.getAttribute('class') == "closebox-icon";
        if (doClose) {
            // do the opposite, we want to show conditions based on this image.
            this.openToggleBox(vboxNode, targetNodeArray, imageNode);
        } else {
            this.closeToggleBox(vboxNode, targetNodeArray, imageNode);
        }
    }
    // Setup the services trees
};

this.updateTargetsMenu = function(names) {
    names.sort();
    var menupopup = document.getElementById("target_textbox_menupopup");
    while (menupopup.firstChild) {
        menupopup.removeChild(menupopup.firstChild);
    }
    var menuitem;
    for each (var name in names) {
        menuitem = document.createElement("menuitem");
        menuitem.setAttribute("label", name);
        menupopup.appendChild(menuitem);
    }
    ["target_textbox_orig", "change_target_button", "revert_target_button"].
        forEach(function(name) {
            document.getElementById(name).setAttribute("hidden", "true");
        });
    document.getElementById("target_textbox").removeAttribute("hidden");
};

this.refreshAppTree = function() {
    this.setupApplicationsTree(false);
};

this.clearTrees = function() {
    for each (var view in widgets.views) {
            if (view.clear) {
                view.clear();
            }
        }
};

// Selection handlers

this.mainTree_OnSelect = function() {
    this.updateStackatoToolbar();
};

this.mainTree_onDoubleClick = function() {
    // There's nothing to do
};

this.mainTree_onKeyPress = function(event) {
    var t = event.originalTarget;
    if (t.localName != "treechildren" && t.localName != 'tree') {
        return false;
    }
    // Special-case some commands, and then look at the keybinding set
    // to determine a command to do.
    if (!(event.shiftKey || event.ctrlKey || event.altKey)) {
        if (this.arrowKeys.indexOf(event.keyCode) >= 0) {
            // Nothing to do but squelch the keycode
            this.updateStackatoToolbar();
        } else if (event.keyCode == event.DOM_VK_ENTER
                   || event.keyCode == event.DOM_VK_RETURN) {
            // What to do?
        } else if (event.keyCode == event.DOM_VK_DELETE) {
            this._application_command(event, "delete", true);
        }
    }
    return true;
};

this.servicesSystemTree_OnSelect = function() {
    //log.debug(">> servicesSystemTree_OnSelect");
};

this.servicesProvisionedTree_OnSelect = function() {
    //log.debug(">> servicesProvisionedTree_OnSelect");
};

this.frameworksTree_OnSelect = function() {
    //log.debug(">> frameworksTree_OnSelect");
};

this.runtimesTree_OnSelect = function() {
    //log.debug(">> runtimesTree_OnSelect");
};

this.targetsTree_OnSelect = function() {
    //log.debug(">> targetsTree_OnSelect");
};

this.onUnload= function() {
    log = ko.logging.getLogger("stackato");
};

///////////////////////////////////////////////////////////////////////
// Context Menu Handlers

this.getStackatoEnv = function(obj) {
    // obj.stackato and obj.path will be quoted if necessary
    // obj.cwd won't be, since it's usually passed as a separate arg.
    var prefs = gko.prefs;
    try {
        obj.stackato = prefs.getStringPref("stackato.location");
    } catch(ex) {
        obj.stackato = null;
    }
    if (!obj.stackato) {
        var prompt = bundle.GetStringFromName("Komodo needs to know where the Stackato executable is located");
        var title = bundle.GetStringFromName("Stackato Configuration");
        var defaultResponse = bundle.GetStringFromName("No");
        var response = gko.dialogs.yesNo(prompt, defaultResponse, null, title);
        if (response == bundle.GetStringFromName('Yes')) {
            prefs_doGlobalPrefs('stackatoItem');
            return;
        }
    }
    obj.stackato = this.quote_if_needed(obj.stackato);
    var sysUtilsSvc = Components.classes['@activestate.com/koSysUtils;1'].getService(Components.interfaces.koISysUtils);
    var osSvc = Components.classes["@activestate.com/koOs;1"].getService(Components.interfaces.koIOs);
    var osPathSvc = Components.classes["@activestate.com/koOsPath;1"].getService(Components.interfaces.koIOsPath);
    var environSvc = Components.classes["@activestate.com/koUserEnviron;1"].getService(Components.interfaces.koIUserEnviron);
    var path = environSvc.get("PATH");
    obj.path = this.quote_if_needed([osPathSvc.dirname(obj.stackato), path].join(osSvc.pathsep));
    try {
        var file = gko.views.manager.currentView.koDoc.file;
        if (file.scheme == "file") {
            obj.cwd = file.dirName;
        } else {
            obj.cwd = null;
        }
    } catch(ex2) {
        log.exception(ex2, "error getting current view dir:");
    }
};

this.quote_if_needed = function quote_if_needed(s) {
    var s1;
    if (/[^\w.:=\-\"\'\\\/]/.test(s)) {
        s1 = '"' + s + '"';
    } else {
        s1 = s;
    }
    return s1;
};

this.getSelectionInfo = function(view, index) {
    var row;
    if (index < 0 || index >= view.rowCount) {
        row = null;
    } else {
        var thisRow = view.getRow(index);
        var mainRow = thisRow.fields.name ? thisRow : view.getMainRow(index);
        row = {
          name: mainRow.fields.name,
          health: mainRow.fields.health,
          num_instances: mainRow.fields.num_instances,
          serviceName: thisRow.fields.serviceName,
          url: thisRow.fields.url,
          env_name: thisRow.fields.env_name,
        };
    }
    return {
      empty: !mainRow,
      fields: row,
      loggedOut: widgets.fields["user_logged_out"].getAttribute("hidden") != "true",
      showingEnvironmentVariables: this.showEnvironmentVariables
    };
};

this.updateStackatoToolbar = function() {
    var view = widgets.views.mainTree;
    var index = view.selection.currentIndex;
    var row;
    var selectionInfo = this.getSelectionInfo(view, index);
    // This should just work.  All we care about is enabling/disabling these.
    var nodeParent = document.getElementById("stackato_apps_toolbar");
    var childNodes = nodeParent.childNodes;
    for (var i = childNodes.length - 1; i >= 0; --i) {
        this._doMainTreeContextMenu(childNodes[i], selectionInfo);
    }
};

this.initMainTreeContextMenu = function(event, menupopup) {
    var clickedNodeId = event.explicitOriginalTarget.id;
    if (clickedNodeId == "mtcm_bindService_popup") {
        return true;
    } else if (clickedNodeId != "mainTree_treechildren") {
        log.debug("Rejecting node " + clickedNodeId);
        return false;
    }
    var row = {};
    widgets.trees.mainTree.treeBoxObject.getCellAt(event.pageX, event.pageY, row, {},{});
    var index = row.value;
    var selectionInfo = this.getSelectionInfo(widgets.views.mainTree, index);
    var childNodes = menupopup.childNodes;
    for (var i = childNodes.length - 1; i >= 0; --i) {
        this._doMainTreeContextMenu(childNodes[i], selectionInfo);
    }
    return true;
};

this._doMainTreeContextMenu = function(menuNode, selectionInfo) {
    var hideIf = menuNode.getAttribute('hideIf');
    var ptn = /^field:(.*?)(?:=(.*))?$/;
    var m, fieldName, fieldValue;
    var fields = selectionInfo.fields;
    if (hideIf && fields) {
        m = ptn.exec(hideIf);
        if (m) {
            fieldName = m[1];
            fieldValue = m[2] || '';
            if (fields[fieldName]) {
                if (fieldValue) {
                    if (fieldValue == fields[fieldName]) {
                        menuNode.setAttribute('collapsed', true);
                        return; // No need to do anything else
                    }
                } else {
                    menuNode.setAttribute('collapsed', true);
                    return; // No need to do anything else
                }
            }
        }
    }
    var hideUnless = menuNode.getAttribute('hideUnless');
    if (hideUnless == "showingEnvironmentVariables"
        && !selectionInfo.showingEnvironmentVariables) {
        menuNode.setAttribute('collapsed', true);
        return; // No need to do anything else
    }
    menuNode.removeAttribute('collapsed');
    var disableNode = false;
    var disableIf = menuNode.getAttribute('disableIf');
    if (disableIf) {
        for each (var disableIfVal in disableIf.split(';')) {
                if (disableIfVal == 'empty') {
                    if (!selectionInfo.fields) {
                        disableNode = true;
                        break;
                    }
                } else if (disableIfVal == 'loggedOut') {
                    if (selectionInfo.loggedOut) {
                        disableNode = true;
                        break;
                    }
                } else if (disableIfVal == "unrecognizedServiceName") {
                    if (!this.handleDatabase(fields.serviceName)) {
                        disableNode = true;
                        break;
                    }
                } else if (disableIfVal == "noBoundApp") {
                    if (!this.getAppForServiceName(fields.serviceName)) {
                        disableNode = true;
                        break;
                    }
                } else if (selectionInfo.fields) {
                    m = ptn.exec(disableIfVal);
                    if (m) {
                        fieldName = m[1];
                        fieldValue = m[2] || '';
                        if (fields[fieldName]) {
                            if (fieldValue) {
                                if (fieldValue == fields[fieldName]) {
                                    disableNode = true;
                                }
                            } else {
                                disableNode = true;
                            }
                        }
                    }
                }
            }
    }
    if (!disableNode) {
        var disableUnless = menuNode.getAttribute('disableUnless');
        if (disableUnless) {
            if (disableUnless == 'empty') {
                if (selectionInfo.fields) {
                    var disableNode = true;
                }
            } else if (selectionInfo.fields) {
                m = ptn.exec(disableUnless);
                if (m) {
                    fieldName = m[1];
                    if (!fields[fieldName]) {
                        disableNode = true;
                    } else {
                        fieldValue = m[2] || '';
                        if (fieldValue && fieldValue != fields[fieldName]) {
                            disableNode = true;
                        }
                    }
                }
            }
        }
    }
    if (disableNode) {
        menuNode.setAttribute('disabled', true);
    } else {
        menuNode.removeAttribute('disabled');
    }
    this._adjustTooltiptextAttribute(menuNode, disableNode);
};

this._adjustTooltiptextAttribute = function(menuNode, disableNode) {
    /**
     * If this node has a disabled_tooltiptext attr, and we're disabling it,
     * set the tooltiptext to it to explain why.
     * Of less obvious use, do same with enabled_tooltiptext for enabled nodes.
     */
    var usePrefix, dropPrefix, useAttr, dropAttr;
    [usePrefix, dropPrefix] = disableNode ? ["dis", "en"] : ["en", "dis"];
    useAttr = usePrefix + "abled_tooltiptext";
    dropAttr = dropPrefix + "abled_tooltiptext";
    if (menuNode.hasAttribute(dropAttr)) {
        menuNode.removeAttribute("tooltiptext");
    }
    if (menuNode.hasAttribute(useAttr)) {
        menuNode.setAttribute("tooltiptext",
                              menuNode.getAttribute(useAttr));
    }
}

this.initProvisionedServicesContextMenu = function(event, menupopup) {
    var row = {};
    widgets.trees.servicesProvisionedTree.treeBoxObject.getCellAt(event.pageX, event.pageY, row, {},{});
    var index = row.value;
    var view = widgets.views.servicesProvisionedTree;
    var selectedRow = ((index < 0 || index > view.rowCount)
                       ? null : view.getRow(index));
    var selectionInfo = {
      empty: !selectedRow,
      loggedOut: widgets.fields["user_logged_out"].getAttribute("hidden") != "true",
      fields: {}
    };
    if (selectedRow) {
        selectionInfo.fields.serviceName = view.getNameAtIndex(index);
    }
    var childNodes = menupopup.childNodes;
    for (var i = childNodes.length - 1; i >= 0; --i) {
        this._doMainTreeContextMenu(childNodes[i], selectionInfo);
    }
    return true;
};

this._applicationCurrentRow = function(event) {
    return widgets.views.mainTree.selection.currentIndex;
};

this._updateApplicationsTable = function() {
    this.wrapCallbackFunction("getApplications",
                              "toggle_applications_button",
                              widgets.views.mainTree,
                              ko.stackato.focus);
};

this.internAppNamePrefset = function(appName) {
    var outerAppPrefs = gStackatoPrefs.getPref("apps");
    var appPrefs;
    if (outerAppPrefs.hasPref(appName)) {
        appPrefs = outerAppPrefs.getPref(appName);
    } else {
        appPrefs = Components.classes["@activestate.com/koPreferenceSet;1"].createInstance();
        outerAppPrefs.setPref(appName, appPrefs);
    }
    return appPrefs;
}

this._getMainRowAndAppName = function(event) {
    var index = this._applicationCurrentRow(event);
    if (index < 0) return [null, null];
    var row = widgets.views.mainTree.getMainRow(index);
    if (!row) {
        log.debug("No row for this item");
        return [row, null];
    }
    var appName = row.fields.name;
    if (!appName) {
        log.debug("No appName for row " + index);
    }
    return [row, appName];
};

this.application_browseCode = function(event) {
    var row, appName;
    [row, appName] = this._getMainRowAndAppName(event);
    if (!row || !appName) return;
    var appPrefs = this.internAppNamePrefset(appName);
    var pwd = null;
    if (appPrefs && appPrefs.hasPref("pwd")) {
        pwd = appPrefs.getStringPref("pwd");
    }
    if (!pwd) {
        pwd =  gko.filepicker.getFolder(null, bundle.GetStringFromName("Where is this project located"));
        if (pwd) {
            appPrefs.setStringPref("pwd", pwd);
        }
    }
    if (!pwd) {
        return;
    }
    gko.views.manager.notify_visited_directory(pwd);
};

this.application_reset_directory = function(event) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var appPrefs = this.internAppNamePrefset(appName);
    if (!appPrefs) {
        ko.dialogs.alert("Komodo/Stackato: Internal Error: can't find prefs for application \"" + appName + '"');
        return;
    }
    var pwd = (appPrefs.hasPref("pwd") ? appPrefs.getStringPref("pwd") : null);
    var newPwd = gko.filepicker.getFolder(pwd, bundle.GetStringFromName("Where is this project located"));
    if (newPwd) {
        appPrefs.setStringPref('pwd', newPwd);
    }
    window.focus();
}

this.application_update = function(event) {
    var row, appName;
    [row, appName] = this._getMainRowAndAppName(event);
    if (!row || !appName) return;
    var data = row.data;
    var pwd = null;
    var window;
    var appPrefs = this.internAppNamePrefset(appName);
    try {
        pwd = appPrefs.getStringPref("pwd");
    } catch(ex) {
    }
    if (!pwd) {
        if (data.pwd) {
            pwd = data.pwd;
        } else {
            var project = gko.projects.manager.currentProject;
            if (project) {
                pwd = project.getFile().dirName;mainTreeSort
            }
            pwd =  gko.filepicker.getFolder(pwd, bundle.GetStringFromName("Where is this project located"));
            if (!pwd) {
                return;
            }
            data.pwd = pwd;
        }
        appPrefs.setStringPref("pwd", pwd);
    }
    
    var command = "update";
    var args = [command, appName, "--path", this.quote_if_needed(pwd), "-n"];
    var terminationCallback = function(retval) {
        log.debug("cmd <" + command + "> finished => " + retval);
        ko.stackato._updateApplicationsTable();
    };
    this.doApplicationCommand(args, terminationCallback);
};

this._application_command = function(event, commandName, prompt) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var actualPrompt = (prompt
                        ? bundle.formatStringFromName("X application Y", [commandName, appName], 2)
                        : null);
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("restart finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand([commandName, appName], terminationCallback,
                              actualPrompt, "toggle_applications_button");
}

this.application_restart = function(event) {
    this._application_command(event, "restart");
};

this.application_start = function(event) {
    this._application_command(event, "start");
};

this.application_stop = function(event) {
    this._application_command(event, "stop");
};

this.application_rename = function(event) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var prompt = bundle.GetStringFromName("New Application Name");
    var label = bundle.GetStringFromName("Name Colon");
    var value = appName;
    var title = prompt;
    var newAppName = ko.dialogs.prompt(prompt, label, value, title);
    if (!newAppName || appName == newAppName) {
        _stackatoWindow.focus();
        return;
    }
    if (widgets.views.mainTree.getNames().indexOf(newAppName) !== -1) {
        gko.statusBar.AddMessage(bundle.formatStringFromName("Stackato name in use X", [newAppName], 1),
                                 bundle.GetStringFromName("Stackato"),
                                 3000, true);
        _stackatoWindow.focus();
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("rename finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand(["rename", appName, newAppName],
                              terminationCallback,
                              null, "toggle_applications_button");
};

this.application_delete = function(event) {
    this._application_command(event, "delete", true);
};

this.application_refresh = function() {
    this._updateApplicationsTable();
};

this.application_add = function(event) {
    var getNewAppFieldResults = {};
    var res = ko.windowManager.openOrFocusDialog(
        "chrome://stackatotools/content/stackatoNewApp.xul",
        "komodo_stackato",
        "chrome,all,close=yes,resizable,dependent=no,modal=yes",
       {gko:gko, 'window':gWindow, ko:ko,
        stackato: this,
         currentTarget: this._target,
        results:getNewAppFieldResults});
    if (!res) {
        // canceled out
        return;
    }
    var this_ = this;
    var terminationCallbackRebuildTable = function(retval) {
        try {
        if (retval) {
            return;
        }
        this_._updateApplicationsTable();
        } catch(ex) {
            log.debug("terminationCallbackRebuildTable: " + ex );
        }
    };
    var terminationCallbackStartService = function(retval) {
        try {
        if (retval) return;
        var args = ['start', getNewAppFieldResults.appname];
        this_.doApplicationCommand(args, terminationCallbackRebuildTable);
        } catch(ex) {
            log.exception(ex, "terminationCallbackStartService: ");
        }
    };
    var terminationCallbackBindService = function(retval) {
        try {
        if (retval) return;
        if (getNewAppFieldResults.appname
            && getNewAppFieldResults.path) {
            var appPrefs = this_.internAppNamePrefset(getNewAppFieldResults.appname);
            appPrefs.setStringPref("pwd", getNewAppFieldResults.path);
        }
        // Update the prefs before doing anything else.
        if (getNewAppFieldResults.provisionedService) {
            var args = ['bind-service',
                        getNewAppFieldResults.provisionedService,
                        getNewAppFieldResults.appname];
            this_.doApplicationCommand(args, terminationCallbackStartService);
        } else {
            terminationCallbackStartService(0);
        }
        } catch(ex) {
            log.exception(ex, "terminationCallbackBindService:");
        }
    };
    var args = ['push'];
    // Names:
    // appname path url mem instances runtime startImmediately
    var required_names = ['appname', 'path', 'url', 'mem'];
    for each (var name in required_names) {
        if (!getNewAppFieldResults[name]) {
            log.debug("results." + name + " not specified");
            return;
        }
    }
    args.push(getNewAppFieldResults.appname,
              "--path", getNewAppFieldResults.path,
              "--url",  getNewAppFieldResults.url);
    if (getNewAppFieldResults.instances) {
        args.push("--instances", getNewAppFieldResults.instances);
    }
    if (getNewAppFieldResults.runtime) {
        args.push("--runtime", getNewAppFieldResults.runtime);
    }
    if (!getNewAppFieldResults.startImmediately) {
        args.push("--no-start");
    }
    this.doApplicationCommand(args, terminationCallbackBindService);
};

this.application_mapUrl = function(event) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var urls = widgets.views.mainTree.getURLsForApp(appName);
    var prompt = bundle.GetStringFromName("New URL To Add");
    var label = bundle.GetStringFromName("URLColon");
    var value = urls.length ? urls[urls.length - 1] : "";
    var title = prompt;
    var newURL = ko.dialogs.prompt(prompt, label, value, title);
    if (!newURL) {
        return;
    } else if (widgets.views.mainTree.getURLs().indexOf(newURL) !== -1) {
        gko.statusBar.AddMessage(bundle.formatStringFromName("URL is already in use X", [newURL], 1),
                                 bundle.GetStringFromName("Stackato"),
                                 3000, true);
        _stackatoWindow.focus();
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("mapURL finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand(["map", appName, newURL],
                              terminationCallback,
                              null, "toggle_applications_button");
};

this.application_unmapUrl = function(event) {
    var index = this._applicationCurrentRow(event);
    if (index < 0) return;
    var thisRow = widgets.views.mainTree.getRow(index);
    var mainRow = thisRow.fields.name ? thisRow : widgets.views.mainTree.getMainRow(index);
    var appName = mainRow.fields.name;
    var url = thisRow.fields.url;
    if (!appName || !url) {
        dump("No "
             + (appName ? "url" : "appname")
             + " for row " + index + "\n");
        _stackatoWindow.focus();
        return;
    }
    var cmd = "unmap";
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug(cmd + " finished => " + retval);
        this_._updateApplicationsTable();
    };
    var prompt = bundle.formatStringFromName("unmap URL X", [url], 1);
    this.doApplicationCommand([cmd, appName, url],
                              terminationCallback,
                              prompt, "toggle_applications_button");
};

this.application_changeInstances = function(event) {
    var mainRow, appName;
    [mainRow, appName] = this._getMainRowAndAppName(event);
    if (!mainRow || !appName) return;
    var numInstances = mainRow.fields.num_instances;
    var prompt = bundle.GetStringFromName("Number of Instances To Run");
    var label = bundle.GetStringFromName("Num Instances");
    var value = numInstances;
    var title = prompt;
    var newNumInstances = ko.dialogs.prompt(prompt, label, value, title);
    if (!newNumInstances || newNumInstances == numInstances) {
        return;
    } else if (!/^\d+$/.test(newNumInstances)) {
        gko.statusBar.AddMessage(bundle.formatStringFromName("Instances must be a number, rejecting X", [newNumInstances], 1),
                                 bundle.GetStringFromName("Stackato"),
                                 3000, true);
        _stackatoWindow.focus();
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("instances finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand(["instances", appName, newNumInstances],
                              terminationCallback,
                              null, "toggle_applications_button");
};

const mem1K = 1024;
const mem1M = mem1K * 1024;
const mem1G = mem1M * 1024;
this.application_changeMemory = function(event) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var this_ = this;
    var handler = {
        setData: function(data) {
            try {
                var memUsedLabel;
                var mem_quota_bytes = data[0].stats.mem_quota;
                if (mem_quota_bytes >= mem1G) {
                    memUsedLabel = Math.round(mem_quota_bytes / mem1G) + "G";
                } else if (mem_quota_bytes >= mem1M) {
                    memUsedLabel = Math.round(mem_quota_bytes / mem1M) + "M";
                } else if (mem_quota_bytes >= mem1K) {
                    memUsedLabel = Math.round(mem_quota_bytes / mem1K) + "K";
                } else {
                    memUsedLabel = mem_quota_bytes + "B";
                }
                var results = {memoryLimit: memUsedLabel};
                var res = ko.windowManager.openOrFocusDialog(
                    "chrome://stackatotools/content/stackatoSelectMemory.xul",
                    "komodo_stackato",
                    "chrome,all,close=yes,resizable,dependent=no,modal=yes",
                   results);
                if (!res || !results.newMemoryLimit || results.newMemoryLimit == results.memoryLimit) {
                    // canceled out
                    return;
                }
                this_.doApplicationCommand(["mem", appName, results.newMemoryLimit],
                                           this_.focus,
                                           null, "toggle_applications_button");
            } catch(ex) {
                gko.dialogs.alert("Komodo/Stackato internal error: " + ex);
            }
        }
    }
    this.wrapCallbackFunction("runCommand",
                              "toggle_applications_button",
                              handler, null, true, /*useJSON*/
                              ["stats", appName, "--json"]);
};

this.application_addEnvVar = function(event) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var prompt = bundle.GetStringFromName("New environment variable__use name-value format");
    var label = bundle.GetStringFromName("Variable and value");
    var value = "";
    var title = bundle.GetStringFromName("Add New Environment Variable");
    var res = ko.dialogs.prompt(prompt, label, value, title);
    if (!res) {
        return;
    }
    if (res.indexOf('=') == -1) {
        res += "=";
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("env_add finished => " + retval);
        this_.getAndUpdateEnvironmentVarsForAppname(view, appName, this_.focus);
    };
    this.doApplicationCommand(["env-add", appName, res],
                              terminationCallback,
                              null, "toggle_applications_button");
};

this.application_removeEnvVar = function(event) {
    var index = this._applicationCurrentRow(event);
    if (index < 0) return;
    var view = widgets.views.mainTree;
    var thisRow = view.getRow(index);
    var mainRow = view.getMainRow(index);
    var appName = mainRow.fields.name;
    if (!appName) {
        dump("No appname for row " + index + "\n");
        return;
    }
    var envVarName = thisRow.fields.env_name;
    if (!envVarName) {
        log.debug("No env var for row " + index);
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("env_del finished => " + retval);
        this_.getAndUpdateEnvironmentVarsForAppname(view, appName, this_.focus);
    };
    var prompt = bundle.formatStringFromName('drop environment variable X',
                                             [envVarName], 1);
    this.doApplicationCommand(["env-del", appName, envVarName],
                              terminationCallback,
                              prompt, "toggle_applications_button");
};

this.bindService_show_services = function(event, menupopup) {
    while (menupopup.firstChild) {
        menupopup.removeChild(menupopup.firstChild);
    }
    var index = this._applicationCurrentRow(event);
    if (index < 0) return false;
    var appName = widgets.views.mainTree.getAppNameForRow(index);
    if (!appName) {
        log.debug("No appname for row " + index);
        return false;
    }
    var currentServiceNames = widgets.views.mainTree.getServicesForApp(appName);
    var menuitem, handler;
    for each (var name in this.provisionedServiceNames) {
            menuitem = document.createElement("menuitem");
            menuitem.setAttribute("label", name);
            menuitem.setAttribute("id", "bindService_" + name);
            if (currentServiceNames.indexOf(name) !== -1) {
                menuitem.setAttribute("disabled", "true");
            } else {
                handler = ("ko.stackato.application_bindService('"
                           + appName
                           + "', '"
                           + name
                           + "');");
                menuitem.setAttribute("oncommand", handler);
            }
            menupopup.appendChild(menuitem);
        }
    return true;
};

this.application_bindService = function(appName, serviceName) {
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("bind-service finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand(["bind-service", serviceName, appName],
                              terminationCallback,
                              null, "toggle_applications_button");
};

this.application_unbindService = function(event) {
    var index = this._applicationCurrentRow(event);
    if (index < 0) return;
    var thisRow = widgets.views.mainTree.getRow(index);
    var mainRow = thisRow.fields.name ? thisRow : widgets.views.mainTree.getMainRow(index);
    var appName = mainRow.fields.name;
    var serviceName = thisRow.fields.serviceName;
    if (!appName || !serviceName) {
        dump("No "
             + (appName ? "serviceName" : "appname")
             + " for row " + index + "\n");
        _stackatoWindow.focus();
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("unbind-service finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand(["unbind-service", serviceName, appName],
                              terminationCallback,
                              null, "toggle_applications_button");
};

this.exploreInDBExplorer = function(event) {
    var index = this._applicationCurrentRow(event);
    if (index < 0) return;
    var thisRow = widgets.views.mainTree.getRow(index);
    var mainRow = thisRow.fields.name ? thisRow : widgets.views.mainTree.getMainRow(index);
    var appName = mainRow.fields.name;
    var serviceName = thisRow.fields.serviceName;
    this.finishExploreInDBExplorer(appName, serviceName);
};

this.finishExploreInDBExplorer = function(appName, serviceName) {
    var dbInfo = this.databasesByAppName[appName][serviceName];
    if (!dbInfo) {
        log.error("No DB Info for "
                  + appName
                  + "/"
                  + serviceName);
        return;
    }
    // So now look to see if this database is in the dbexplorer....
    var dbVendorName = this.getDBVendor(serviceName);
    if (!dbVendorName
        || !this._dbExplorerDatabases[dbVendorName.toLowerCase()]) {
        if (!dbVendorName) {
            ko.dialogs.alert(bundle.formatStringFromName("Cant figure out type of database X", [dbInfo.databaseTypeName], 1));
        } else {
            ko.dialogs.alert(bundle.formatStringFromName("Database X isnt supported in dbexplorer yet", [dbVendorName], 1));
        }
        return;
    }
    var dbExplorerDBName = this._dbExplorerDatabases[dbVendorName] || dbVendorName;
    var dbNames, dbValues;
    var dbvm = gko.dbexplorer.viewManager;
    var dbview = dbvm.view;
    var dbRowNums = [];
    var thisConn;
    for (var i = 0; i < dbview.rowCount; i++) {
        if (dbview.getNodeType(i) == "connection") {
            dbNames = {};
            dbValues = {};
            dbview.getConnectionInfo(i, {}, dbNames, {}, dbValues);
            dbNames = dbNames.value;
            dbValues = dbValues.value;
            thisConn = {};
            for (var j = 0; j < dbNames.length; j++) {
                thisConn[dbNames[j]] = dbValues[j];
            }
            if (thisConn.db_type == dbExplorerDBName
                && thisConn.hostname == dbInfo.host
                && (!thisConn.username || thisConn.username == dbInfo.user)
                && (!thisConn.port || thisConn.port == dbInfo.port)) {
                // We found it, so show the tab, highlight the row, and get out of here.
                gWindow.focus();
                gko.uilayout.toggleTab("dbexplorer_viewbox", false, false);
                dbview.selection.currentIndex = i;
                if (dbview.isContainer(i)
                    && !dbview.isContainerOpen(i)
                    && !dbview.isContainerEmpty(i)) {
                    dbview.toggleOpenState(i);
                }
                return;
            }
        }
    }
    // Now go add the entry.
    var uri = "chrome://dbexplorer/content/addConnection.xul";
    var windowType = "kodbexplorer:addConnection";
    var features = [
        "menubar=no",
        "toolbar=no",
        "location=no",
        "status=yes",
        "scrollbars=yes",
        "resizable=yes",
        "minimizable=yes",
        "chrome=yes",
        "modal=yes"
    ].join(",");
    var fields = {
      dbTypeName: dbExplorerDBName,
      hostname: dbInfo.host,
      port: dbInfo.port,
      username: dbInfo.user,
      password: dbInfo.passwd,
      showPassword: !!dbInfo.passwd,
      hasPassword: !!dbInfo.passwd,
    };
    var dbResultObj = {ko:ko,
                       view:dbview,
                       rowNum:-1, // Not used when adding a connection
                       operation:'add',
                       fields: fields
    };
    var win = window.openDialog(uri, windowType, features, dbResultObj);
    if (dbResultObj.result) {
        // Show the dbx tab.
        // Even better, open up the new window.
        gWindow.focus();
    }
};

this.application_showInfo = function(event, infoName) {
    var appName = this._getMainRowAndAppName(event)[1];
    if (!appName) return;
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug(infoName + " finished => " + retval);
        this_._updateApplicationsTable();
    };
    this.doApplicationCommand([infoName, appName],
                              terminationCallback,
                              null, "toggle_applications_button");
};

this.application_showFiles = function(event, infoName) {
    var row, appName;
    [row, appName] = this._getMainRowAndAppName(event);
    if (!row || !appName) return;
    var numInstances = mainRow.fields.num_instances;
    var results = {};
    var res = ko.windowManager.openOrFocusDialog(
        "chrome://stackatotools/content/stackatoFilesArgs.xul",
        "komodo_stackato",
        "chrome,all,close=yes,resizable,dependent=no,modal=yes",
       {gko:gko, 'window':gWindow, ko:ko,
        stackato: this,
        numInstances:numInstances,
        results:results});
    if (!res) {
        // canceled out
        return;
    }
    var args = ["files", appName];
    if (results.path) {
        args.push(results.path);
    }
    if (results.instanceNum && results.instanceNum != "*") {
        args.push("--instance", results.instanceNum);
    }
    if (results.showAll) {
        args.push("--all");
    }
    this.doApplicationCommand(args, null, null, null);
};

this.doApplicationCommand = function(args, terminationCallback, prompt, iconID) {
    if (typeof(prompt) == "undefined") prompt = null;
    if (typeof(iconID) == "undefined") iconID = null;
    var obj = {};
    if (prompt) {
        var prompt2 = bundle.formatStringFromName("Please confirm that Stackato should X",
                                                  [prompt], 1);
        var response = "Cancel";
        var text = null;
        var title = bundle.GetStringFromName("Confirm possible destructive action");
        var res = ko.dialogs.okCancel(prompt2, response, text, title);
        if (res != "OK") {
            return;
        }
    }                                      
    this.getStackatoEnv(obj);
    if (!obj.stackato || !obj.path) {
        ko.dialogs.alert(bundle.formatStringFromName("Bailing out of X",
                                                     args, 1));
        return;
    }
    args.unshift(obj.stackato);
    args.push("-n");
    var command = args.join(" ");
    var env = "PATH=" + obj.path;
    // yikes, Komodo has only one terminal window...
    var toggleButtonImage, prevClass, actualCallback;
    var callback2 = function(retval) {
        if (terminationCallback) {
            terminationCallback(retval);
        }
    };
    if (iconID) {
        toggleButtonImage = document.getElementById(iconID);
        prevClass = toggleButtonImage.getAttribute("class");
        toggleButtonImage.setAttribute("class", "async_operation");
        actualCallback = function(retval) {
            toggleButtonImage.setAttribute("class", prevClass);
            callback2(retval);
        }
    } else {
        actualCallback = callback2;
    }
    if (!gko.run.runCommand(gWindow, command, null, env, false,
                            false, false, "command-output-window",
                            /*parseOutput*/ false,
                            /*parseRegex*/ false,
                            /*showParsedOutputList*/ false,
                            /*name*/ null,
                            /*clearOutputWindow*/ true,
                            actualCallback)) {
        actualCallback(-1);
    }
};


this._provisionedServicesCurrentRow = function(event) {
    return widgets.views.servicesProvisionedTree.selection.currentIndex;
};

this._updateProvisionedServicesTable = function() {
    var this_ = this;
    var callback = function() {
        this_.provisionedServiceNames = widgets.views.servicesProvisionedTree.getNames();
        this_.focus();
    };
    this.wrapCallbackFunction("getServices",
                              "toggle_provisioned_services_button",
                              this.servicesTreeManager,
                              callback);
};

this.provisionedServices_createService = function(event) {
    var results = {};
    var res = ko.windowManager.openOrFocusDialog(
        "chrome://stackatotools/content/stackatoNewService.xul",
        "komodo_stackato",
        "chrome,all,close=yes,resizable,dependent=no,modal=yes",
       {gko:gko, 'window':gWindow, ko:ko,
        stackato: this,
        results:results});
    if (!res) {
        // canceled out
        return;
    }
    if (!results.baseServiceName || !results.provisionedServiceName) {
        return;
    }
    if (this.systemServiceNames.indexOf(results.baseServiceName) === -1) {
        gko.statusBar.AddMessage(bundle.formatStringFromName("Unknown service X",
                                                             [results.baseServiceName], 1),
                                 bundle.GetStringFromName("Stackato"),
                                 3000, true);
        return;
    }
    if (this.provisionedServiceNames.indexOf(results.provisionedServiceName) !== -1) {
        gko.statusBar.AddMessage(bundle.formatStringFromName("Service already defined X",
                                 [results.provisionedServiceName], 1),
                                 bundle.GetStringFromName("Stackato"),
                                 3000, true);
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("create-service finished => " + retval);
        this_._updateProvisionedServicesTable();
    };
    this.doApplicationCommand(["create-service",
                               results.baseServiceName,
                               results.provisionedServiceName],
                              terminationCallback,
                              null, "toggle_provisioned_services_button");
};

this.provisionedServices_deleteService = function(event) {
    var index = this._provisionedServicesCurrentRow(event);
    if (index < 0) return;
    var serviceName = widgets.views.servicesProvisionedTree.getNameAtIndex(index);
    if (!serviceName) {
        log.debug("No serviceName for row " + index);
        return;
    }
    var this_ = this;
    var terminationCallback = function(retval) {
        log.debug("delete-service finished => " + retval);
        this_._updateProvisionedServicesTable();
    };
    var prompt = bundle.formatStringFromName('delete provisioned service X',
                                             [serviceName], 1);
    this.doApplicationCommand(["delete-service", serviceName],
                              terminationCallback,
                              prompt, "toggle_provisioned_services_button");
};

this.provisionedServices_exploreInDBExplorer = function(event) {
    var index = this._provisionedServicesCurrentRow(event);
    if (index < 0) return;
    var serviceName = widgets.views.servicesProvisionedTree.getNameAtIndex(index);
    if (!serviceName) {
        log.debug("No serviceName for row " + index);
        return;
    }
    var appName = this.getAppForServiceName(serviceName);
    if (!appName) {
        log.debug("No app for provisioned service " + serviceName);
        return;
    }
    this.finishExploreInDBExplorer(appName, serviceName);
};

this.getAppForServiceName = function(serviceName) {
    return widgets.views.mainTree.getAppForServiceName(serviceName);
};

this.updateProvisionedServiceNames = function(callback, iconID) {
    try {
        this.refreshProvisionedServicesTree(callback, iconID);
    } catch(ex) {
        log.exception(ex, "**************** updateProvisionedServiceNames: ");
    }
};

}).apply(ko.stackato);
