// Copyright (c) 2000-2011 ActiveState Software Inc.
// See the file LICENSE.txt for licensing information.

var gObj;
var widgets = {};
var appNameChanged_ID = 0;
var updateEnabledOK_ID = 0;
var updateEnabledAddService_ID = 0;
var urlExplicitlyUpdated = false;

var bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
    .getService(Components.interfaces.nsIStringBundleService)
    .createBundle("chrome://stackatotools/locale/stackato.properties");

function fillMenu(menupopup, names, defaultValue) {
  while (menupopup.firstChild) {
    menupopup.removeChild(menupopup.firstChild);
  }
  var menuitem;
  for each (var name in names) {
      menuitem = document.createElement("menuitem");
      menuitem.setAttribute("label", name);
      menupopup.appendChild(menuitem);
  }
  menupopup.parentNode.value = defaultValue;
}

function onLoad() {
  gObj = window.arguments[0];
  for each (var name in ["appname", "path", "url", "numInstances",
                         "memoryLimit", "startImmediately",
                         "runtime_menupopup", "provisionedService_menupopup",
                         "systemServices_menupopup",
                         "runtime", "provisionedServices",
                         "new_provisioned_service_name",
                         "new_provisioned_button",
                         "system_services"
                         ]) {
          widgets[name] = document.getElementById(name);
      }
  fillMenu(widgets.runtime_menupopup,
           gObj.stackato.runtimeNames, "");
  fillMenu(widgets.provisionedService_menupopup,
           gObj.stackato.provisionedServiceNames, "");
  fillMenu(widgets.systemServices_menupopup,
           gObj.stackato.systemServiceNames, "");
  updateEnabledOK();
  for each (var name in ["appname", "path", "url", "numInstances",
                         "memoryLimit",]) {
          widgets[name].addEventListener("keypress", queueUpdateEnabledOK, false);
      }
  widgets.appname.addEventListener("keypress", wrapOnAppNameUpdated, false);
  widgets.url.addEventListener("keypress", onUrlUpdated, false);

  widgets.new_provisioned_service_name.addEventListener("keypress",
      queueUpdateEnabledAddService, false);
}

function onShutdown() {
  if (updateEnabledOK_ID) {
      clearTimeout(updateEnabledOK_ID);
  }
  for each (var name in ["appname", "path", "url", "numInstances",
                         "memoryLimit",]) {
          widgets[name].removeEventListener("keypress", queueUpdateEnabledOK, false);
      }
  widgets.new_provisioned_service_name.removeEventListener("keypress", queueUpdateEnabledAddService, false);
}

function onOK() {
  var results = gObj.results;
  // Names:
  // appname path url mem instances runtime startImmediately
  results.appname = widgets.appname.value;
  results.path = widgets.path.value;
  results.url = widgets.url.value;
  results.instances = widgets.numInstances.value;
  results.mem = widgets.memoryLimit.value;
  results.runtime = widgets.runtime.value;
  results.startImmediately = widgets.startImmediately.checked;
  var selectedItem = widgets.provisionedServices.selectedItem;
  if (selectedItem && selectedItem.label) {
      results.provisionedService = selectedItem.label;
  } else {
      results.provisionedService = null;
  }
  onShutdown();
  return true;
}
function onCancel() {
  onShutdown();
  return true;
}

function browsePath() {
    var prompt = bundle.GetStringFromName("The directory containing the application");
    var res = gObj.gko.filepicker.getFolder(widgets.path.value, prompt);
    if (res) {
        widgets.path.value = res;
    }
    window.focus();
}

function generateName_aux(serviceName) {
    // Favor certain letters in suggested names over others.
    var letters = "aaaaabbbbbcccccdddddeeffffgggghhhhhijjjkkklmmmnnnopppqqqrrrssstttuuvvvwwxxxyzzz";
    var digits = "0123456789";
    var rand = Math.random
    var results = [serviceName, "-"];
    results.push(letters[Math.floor(Math.random() * letters.length)]);
    for (var i = 4; i > 0; i--) {
        results.push(digits[Math.floor(Math.random() * digits.length)]);
    }
    return results.join("");
}

function generateName(serviceName) {
    var existingServiceNames = gObj.stackato.provisionedServiceNames;
    var candidate;
    for (var i = 10; i > 0; i--) {
        candidate = generateName_aux(serviceName);
        if (!~existingServiceNames.indexOf(candidate)) {
            return candidate;
        }
        dump("Have collision with " + candidate + ", try another\n");
    }
    return candidate;
}

function onMenuitemChanged(menulist) {
    var textbox = widgets.new_provisioned_service_name;
    var serviceName = menulist.selectedItem.label;
    textbox.value = generateName(serviceName);
    updateEnabledAddService();
    textbox.select();
}

function queueUpdateEnabledOK() {
    if (updateEnabledOK_ID) {
        clearTimeout(updateEnabledOK_ID);
    }
    updateEnabledOK_ID = setTimeout(updateEnabledOK, 400);
}

function onUrlUpdated() {
    urlExplicitlyUpdated = true;
    widgets.url.removeEventListener("keypress", onUrlUpdated, false);
    widgets.appname.removeEventListener("keypress", wrapOnAppNameUpdated, false);
}

function wrapOnAppNameUpdated() {
    // Let the field get updated.
    if (appNameChanged_ID) {
        clearTimeout(appNameChanged_ID);
    }
    appNameChanged_ID = setTimeout(onAppNameUpdated, 100);
}

function onAppNameUpdated() {
    appNameChanged_ID = 0;
    var currentURL = widgets.url.value;
    var appName = widgets.appname.value;
    if (appName.length === 0) {
        widgets.url.value = "";
    } else {
        var target = gObj.currentTarget.substr(gObj.currentTarget.indexOf("."));
        if (!currentURL || currentURL.indexOf(target) >= 0) {
            widgets.url.value = appName + target;
        }
    }
}

function queueUpdateEnabledAddService() {
    if (updateEnabledAddService_ID) {
        clearTimeout(updateEnabledAddService_ID);
    }
    updateEnabledAddService_ID = setTimeout(updateEnabledAddService, 400);
}


function updateEnabledAddService() {
    var disableAddService;
    if (!widgets.new_provisioned_service_name.value) {
        disableAddService = true;
    } else if (!widgets.system_services.selectedItem
               || !widgets.system_services.selectedItem.label) {
        disableAddService = true;
    } else {
        disableAddService = false;
    }
    widgets.new_provisioned_button.disabled = disableAddService;
}

function updateEnabledOK() {
    try {
    var disableOk = false;
    for each (var name in ["appname", "path", "url", "numInstances",
                           "memoryLimit"]) {
            if (!widgets[name].value) {
                disableOk = true;
                break;
            }
        }
    if (!disableOk) {
        disableOk = (!widgets.runtime.selectedItem
                     || !widgets.runtime.selectedItem.label);
    }
    document.documentElement.getButton("accept").disabled = disableOk;
    } catch(ex) {
        dump("updateEnabledOK: " + ex + "\n");
    }
}

function addProvisionedService() {
    var textbox = widgets.new_provisioned_service_name;
    var serviceName = widgets.system_services.selectedItem.label;
    var provisionedServiceName = textbox.value;
    var outerCallback = function() {
        fillMenu(widgets.provisionedService_menupopup,
                 gObj.stackato.provisionedServiceNames,
                 provisionedServiceName);
    };
    var handleTarget = {
      setData: function(data) {
            // Update the list of service names
            gObj.stackato.updateProvisionedServiceNames(outerCallback,
                                                        "toggle_provisioned_services_button");
            // Update the selected provisioned database
        }
    };
    gObj.stackato.wrapCallbackFunction("runCommand",
                                       null,
                                       handleTarget,
                                       null,
                                       false,
                                       ["create-service", serviceName, provisionedServiceName]);
};
                                   
